const {promisify} = require('util');
const builder = require('xmlbuilder');
const checker = require('license-checker');
const getLicenses = promisify(checker.init);
const fs = require('fs');
const readFileAsync = promisify(fs.readFile);
const writeFileAsync = promisify(fs.writeFile);
const existsAsync = promisify(fs.exists);
const parseString = promisify(require('xml2js').parseString);
const path = require('path');

/**
 * @type {boolean} true for windows os.
 */
const isWindowsOs = (process.platform + "").indexOf("win") === 0;
/**
 * Generates the XML file for the given components.
 * @param components The array of components.
 */
async function genXml(components) {
  let root = builder.create('third-party');
  components.forEach(comp => genComponent(root, comp));
  return root.end({ pretty: true });
}


/**
 * If multiple licenses choose one that is not GPL.
 * @param licenses a license or array of licenses.
 * @return a license
 */
function chooseLicense(licenses) {
  if (!Array.isArray(licenses)) {
    licenses = [licenses];
  }
  // Dual licensing - avoid GPL
  let license = licenses
    .filter(lic => !lic.startsWith("GPL"))[0]
  if (!license) {
    throw new Error("Unsuitable license: " + license);
  }
  return license;
}

/**
 * Guess the license URL from a component details.
 *
 * @param component The component descriptor.
 * @return The license URL.
 */
function guessLicenseUrl(component) {
  var licenseUrl = component.licenseUrl;
  if (!licenseUrl && component.path && component.licenseFile && component.repository) {
    var licenseRelPath = 
        component.licenseFile.substring(component.path.length);
    licenseUrl = component.repository + '/blob/master' + 
        licenseRelPath.replace('\\', '/');
  }
  return licenseUrl;
}


/**
 * Generates the XML fragment for a component.
 * @param root The root element.
 * @param component The component descriptor.
 */
function genComponent(root, component) {
  var license = chooseLicense(component.licenses);
  var licenseUrl = guessLicenseUrl(component);

  var attrs = {id: component.id, product: 'content-fusion'};
  if (component.draft) {
    attrs.draft = true;
  }
  root.ele('component', attrs)
    .ele('name', component.name).up()
    .ele('version', component.version).up()
    .ele('project-info')
      .ele('about', component.description).up()
      .ele('licensor', component.publisher).up()
      .ele('start-year', component.startYear).up()
      .ele('project-url', component.repository).up()
    .up()
    .ele('license')
      .ele('type', license).up()
      .ele('content', component.licenseContent).up()
      .ele('url', licenseUrl).up()
    .up()
}

/**
 * Parse a JS file bundled with browserify (with full paths enabled)
 * and extracts all package.json file paths.
 * @param appPath Path to the JS file.
 * @returns {string[]} Sorted list of package.json file paths.
 */
function getPackageJsonFilesList(appPath) {
  let files = [];
  let app = fs.readFileSync(appPath, {encoding: 'utf8'});
  let regexp;
  if (isWindowsOs) {
    regexp = new RegExp('"([^"]+node_modules\\\\\\\\[^\\\\]*)[^"]+"', 'g');
  } else {
    regexp = new RegExp('"([^"]+node_modules/[^/]*)[^"]+"', 'g');
  }

  let match = regexp.exec(app);
  while(match) {
    let nodeModuleLocation = match[1];

    var packageFolders = [];
    if (nodeModuleLocation.match(/@[^/\\]+$/)) {
      // Nodejs modules that start with @ are official packages that are namespaced.
      // So we need to go down one directory.
      fs.readdirSync(nodeModuleLocation).forEach(folder => {
        if (isWindowsOs) {
          packageFolders.push(match[1] + '\\\\' + folder);
        } else {
          packageFolders.push(match[1] + '/' + folder);
        }
      })
    } else {
      packageFolders = [nodeModuleLocation];
    }

    packageFolders.forEach(packageFolder => {
      let packageJson;
      if (isWindowsOs) {
        packageJson = packageFolder + "\\\\package.json";
      } else {
        packageJson = packageFolder + "/package.json";
      }

      if (files.indexOf(packageJson) === -1) {
        files.push(packageJson)
      }
    });

    match = regexp.exec(app);
  }
  files = files.sort();
  return files;
}

/**
 * Return the ID of a component based on the package.json file.
 *
 * @param {string} packageFile - Path to a package.json file to be parsed.
 *
 * @return {string} The component ID in a `name@version` format.
 */
async function getComponentId(packageFile){
  let jsonFile = await readFileAsync(packageFile, {encoding: 'utf8'});
  let json = JSON.parse(jsonFile);
  return `${json.name}@${json.version}`;
}


/**
 * Returns the IDs of the dependencies bundled by browserify.

 * @param appJsPath Path to the browserify bundle.

 * @return the list of ids in a `name@version` format.
 */
async function getBundledDependeciesIds(appJsPath) {
	let files = getPackageJsonFilesList(appJsPath);
	return Promise.all(files.map(getComponentId));
}

/**
 * Flattens an array of arrays and converts to a Set.
 * @param array the array to flatten.
 * @return The flattened set of values.
 */
function flattenToSet(array) {
  let flat = array.reduce((acc, val) => acc.concat(val), []);
  return new Set(flat);
}


/**
 * Sort an array of components by ID.
 * @param {Array<{id: string}>} components
 */
function sortById(components) {
  components.sort(({id: id1}, {id: id2}) => {
    if (id1 < id2) {
      return -1;
    }
    if (id2 < id1) {
      return 1;
    }
    return 0;
  })
}

/**
 * Remove falsy fields from an object.
 *
 * @param obj An object.
 *
 * @return The object without falsy fields.
 */
function removeFalsy(obj) {
  let newObj = {};
  Object.keys(obj).forEach((prop) => {
    if (obj[prop]) { 
      newObj[prop] = obj[prop]; 
    }
  });
  return newObj;
}


/**
 * Load the components from an existing XML file.
 *
 * @param {string} dbPath The path to the XML file.
 * @return the list of descriptors.
 */
async function loadComponents(dbPath) {
  if (!await existsAsync(dbPath)) {
    return [];
  }

  let content = await readFileAsync(dbPath, {encoding: 'utf-8'});
  let json = await parseString(content);

  if (json) {
    let components = json['third-party'].component;
    if (components) {
      return components.map(comp => {
        let projInfo = comp['project-info'][0];
        let licenseInfo = comp['license'][0];

        return removeFalsy({
          id: comp.$.id,
          draft: comp.$.draft,
          name: comp.name[0],
          version: comp.version[0],
          description: projInfo.about[0],
          publisher: projInfo.licensor[0],
          startYear: projInfo['start-year'][0],
          url: projInfo['project-url'][0],
          repository: projInfo['project-url'][0],
          licenses: licenseInfo.type[0],
          licenseContent: licenseInfo.content[0],
          licenseUrl: licenseInfo.url[0]
        });
      });
    }
  }
  return [];
}

/**
 * @param existingComponents The list of existing components.
 * @param components The currnet components.
 *
 * @return {boolean} true if the licenses are up to date.
 */
function areLicensesUpToDate(existingComponents, components) {
  var draftTrueComp = existingComponents.find(comp => comp.draft);
  if (draftTrueComp) {
    console.warn("There are components with @draft=true ", draftTrueComp);
    return false;
  }

  let oldIds = existingComponents.map(comp => `${comp.name}@${comp.version}`);
  let newIds = components.map(comp => `${comp.name}@${comp.version}`);
  let idsMatch = JSON.stringify(oldIds) === JSON.stringify(newIds);

  if (!idsMatch) {
    console.warn("There are new or extra components declared in third-party-components.xml.");
    console.warn("Component IDs currently used:                         ", newIds);
    console.warn("Component IDs declared in third-party-components.xml: ", oldIds);
  }

  return idsMatch;
}


/**
 * Create a list of descriptors for new components, some of which being drafts.
 *
 * @param drafts The array of draft descriptors for new components. 
 * @param existingComponents The array of exising components descriptors.
 *
 * @return A list of draft component descriptors.
 */
function mergeDrafts(drafts, existingComponents) {
  let existingComponentsById = new Map(existingComponents.map(d => [d.id, d]));
  let existingComponentsByName = new Map(existingComponents.map(d => [d.name, d]));

  return drafts.map(draft => {
    // The component was already used - preserve its details.
    let existingComponent = existingComponentsById.get(draft.id);
    if (existingComponent) {
      return existingComponent;
    }

    // The component was already used at a different version - use that as 
    // a starting point.
    let compWithSameName = existingComponentsByName.get(draft.name);
    if (compWithSameName) {
      return Object.assign({draft: true}, compWithSameName, {id: draft.id, version: draft.version});
    }

    // The component is new - use the draft.
    return draft;
  });
}


/**
 * Guesses the license content of a component.
 *
 * @param component The descriptor of a component.
 */
async function guessLicenseContent(component) {
  var content = component.licenseContent;
  if (!content && component.licenseFile) {
    content = await readFileAsync(component.licenseFile, {encoding: 'utf8'});
    component.licenseContent = content;
  }
  return component;
}


/**
 * Create draft descriptors for dependencies using the `license-checker` library. 
 *
 * @param {string} projectPath The path to the project folder.
 * @param {boolean} production true if we should check only production deps.
 *
 * @return an array of draft descriptors.
 */
async function createDrafts(projectPath, production) {
  // Custom format for license-checker
  const customFormat = {
    id: "",
    name: "",
    version: "",
    repository: "",
    publisher: "",
    url: "",
    licenses: "",
    licenseFile: "",
    license: "",
    description: ""
  };

  let draftsCustom = await getLicenses({
    start: projectPath,
    production: production,
    customFormat
  });

  // Ask also for default params since if we specify the format, some
  // details will not be included
  let draftsDefault = await getLicenses({
    start: projectPath,
    production: production
  });

  return Object.keys(draftsCustom)
    .filter(id => {
      var info = draftsCustom[id];
      return info.license !== "OXYGEN EULA"
    })
    .map(id => {
    var info = draftsCustom[id];
    return Object.assign({}, draftsDefault[id], {
      version: info.version,
      name: info.name,
      description: info.description,
      id: `${info.name}@${info.version}`,
      draft: true
    });
  });
}

/**
 * Generates the third-party components file for a given project.
 * @param options.path The path of the project
 * @param options.browserifyBundles An array with the paths to the browserify bundles.
 */
exports.genThirdPartyComponentsXml = async function genThirdPartyComponentsXml(options) {
  options = options || {};
  let projectPath = options.path || '.';
  let xmlFilePath = path.resolve(projectPath, 'third-party-components.xml');
  let browserifyBundles = options.browserifyBundles || [];

	let isBrowserify = browserifyBundles.length !== 0;
  // Browserify bundles also dev-dependencies.
	let production = !isBrowserify;

  let existingComponents = await loadComponents(xmlFilePath);
  let drafts = await createDrafts(projectPath, production);

  let newComponents = mergeDrafts(drafts, existingComponents);

	var usedComponents;
  if (isBrowserify) {
		let idsPerBundle = await Promise.all(browserifyBundles.map(getBundledDependeciesIds));
		let ids = flattenToSet(idsPerBundle);
		usedComponents = newComponents.filter(comp => ids.has(comp.id));
  } else {
		usedComponents = newComponents;
  }

  // Try to read the license content if not already available
  usedComponents = await Promise.all(usedComponents.map(guessLicenseContent))
  sortById(usedComponents);

  if (areLicensesUpToDate(existingComponents, usedComponents)) {
    return true;
  }

  var xmlContent = await genXml(usedComponents);
  await writeFileAsync(xmlFilePath, xmlContent, {encoding: 'utf8'});

  return false;
};

